/*
 * RFCNB (RFC1001/RFC1002) NEtBIOS implementation
 *
 * Version 1.0
 * RFCNB IO Routines ...
 *
 * Copyright (C) Richard Sharpe 1996
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"

static void dump_pkt (const struct RFCNB_Pkt *pkt);

int RFCNB_Timeout     = 0;  /* Timeout in seconds ... (0 is forever) */
int RFCNB_errno       = 0;
int RFCNB_saved_errno = 0;

ProtocolPrinter Prot_Print_Routine = NULL; /* Pointer to print routine */

const struct SMB_protocol_list SMB_protocols[] = {
            { SMB_P_Core,       "PC NETWORK PROGRAM 1.0"  },
            { SMB_P_CorePlus,   "MICROSOFT NETWORKS 1.03" },
            { SMB_P_DOSLanMan1, "MICROSOFT NETWORKS 3.0"  },
            { SMB_P_DOSLanMan1, "DOS LANMAN1.0"           },
            { SMB_P_LanMan1,    "LANMAN1.0"               },
            { SMB_P_DOSLanMan2, "DOS LM1.2X002"           },
            { SMB_P_LanMan2,    "LM1.2X002"               },
            { SMB_P_LanMan2_1,  "DOS LANMAN2.1"           },
            { SMB_P_LanMan2_1,  "LANMAN2.1"               },
            { SMB_P_NT1,        "Samba"                   },
            { SMB_P_NT1,        "NT LM 0.12"              },
            { SMB_P_NT1,        "NT LANMAN 1.0"           },
            { SMB_P_Unknown,    NULL                      }
          };

/*
 * Text strings for the error responses
 */
static const char *error_strings[] = {

  "RFCNBE_OK: Routine completed successfully.",
  "RFCNBE_NoSpace: No space available for a malloc call.",
  "RFCNBE_BadName: NetBIOS name could not be translated to IP address.",
  "RFCNBE_BadRead: Read system call returned an error. Check errno.",
  "RFCNBE_BadWrite: Write system call returned an error. Check errno.",
  "RFCNBE_ProtErr: A protocol error has occurred.",
  "RFCNBE_ConGone: Connection dropped during a read or write system call.",
  "RFCNBE_BadHandle: Bad connection handle passed.",
  "RFCNBE_BadSocket: Problems creating socket.",
  "RFCNBE_ConnectFailed: Connection failed. See errno.",
  "RFCNBE_CallRejNLOCN: Call rejected. Not listening on called name.",
  "RFCNBE_CallRejNLFCN: Call rejected. Not listening for called name.",
  "RFCNBE_CallRejCNNP: Call rejected. Called name not present.",
  "RFCNBE_CallRejInfRes: Call rejected. Name present, but insufficient resources.",
  "RFCNBE_CallRejUnSpec: Call rejected. Unspecified error.",
  "RFCNBE_BadParam: Bad parameters passed to a routine.",
  "RFCNBE_Timeout: IO Operation timed out ..."
};

#if defined(__DJGPP__)
static void RFCNB_alarm (int sig)
{
  psignal (sig, "RFCNB timeout");
}
#else
  #define alarm(x) ((void)0)
#endif

/*
 * Set timeout value and setup signal handling
 */
int RFCNB_Set_Timeout (int seconds)
{
  RFCNB_Timeout = seconds;

#if defined(__DJGPP__)
  if (RFCNB_Timeout > 0)
  {
    struct sigaction inact, outact;

    inact.sa_handler = RFCNB_alarm;
    sigemptyset (&inact.sa_mask);
    inact.sa_flags = 0;         /* Don't restart */

    if (sigaction (SIGALRM, &inact, &outact) < 0)
       return (-1);
  }
#endif
  return (0);
}

/*
 * Discard the rest of an incoming packet as we do not have space for it
 * in the buffer we allocated or were passed ...
 */
static int DiscardRest (const struct RFCNB_Con *con, int len)
{
  char temp[100];
  int  rest = len;

  SMB_DEBUG (2, ("DiscardRest: length %d\n", len));

  while (rest > 0)
  {
    len = read_s (con->fd, temp, min(rest, sizeof(temp)));

    if (len <= 0)   /* Error so return */
    {
      if (len < 0)
           RFCNB_errno = RFCNBE_BadRead;
      else RFCNB_errno = RFCNBE_ConGone;
      RFCNB_saved_errno = errno;
      return (-1);
    }
    rest -= len;
  }
  return (0);
}


/*
 * Send an RFCNB packet to the connection.
 *
 * We just send each of the blocks linked together. Typically the RFCNB_hdr is in
 * 'pkt-data' and the upper-layer SMB-stuff is in 'pkt->next->data'.
 * Check nmap's nselib/smb.lua for details.
 */
int RFCNB_Put_Pkt (const struct RFCNB_Con *con, const struct RFCNB_Pkt *pkt, int len)
{
  const  struct RFCNB_Pkt *p = pkt;
  struct iovec io_list[10];
  int    i, tot_len;

  /* Try to send the data ... We only send as many bytes as len claims.
   * We should try to stuff it into an IOVEC and send as one write.
   */
  for (i = tot_len = 0; p && i < DIM(io_list); p = p->next, i++)
  {
    io_list[i].iov_len  = p->len;
    io_list[i].iov_base = p->data;
    tot_len += p->len;
  }

  SMB_DEBUG (2, ("frags %d, len %d, tot_len %d\n", i, len, tot_len));

  len = writev_s (con->fd, io_list, i);
  if (len < tot_len)
  {
    if (errno == EINTR || errno == ETIMEDOUT)
         RFCNB_errno = RFCNBE_Timeout;
    else RFCNB_errno = RFCNBE_BadWrite;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  RFCNB_Print_Pkt (stdout, "sent", pkt, len);
  return (len);
}

/*
 * Read an RFCNB packet off the connection.
 *
 * We read the first 4 bytes, that tells us the length, then read the
 * rest. We should implement a timeout, but we don't just yet
 */
int RFCNB_Get_Pkt (const struct RFCNB_Con *con, struct RFCNB_Pkt *pkt, int len)
{
  struct RFCNB_Pkt *pkt_frag;
  struct RFCNB_Hdr  hdr;
  int    read_len = 0, pkt_len, rc = 0;
  int    more, this_time, offset;
  int    frag_len, this_len;

  SMB_ASSERT (pkt && pkt->data);

  /* Read that header straight into the buffer
   */
  if (len < sizeof(hdr))
  {
    SMB_DEBUG (1, ("short packet: %s\n", strerror(errno)));
    RFCNB_errno = RFCNBE_BadParam;
    return (-1);
  }

#if 0
  /* rewrite this to use non-blocking reads with select */
#endif

  /* We discard keep alives here ...
   */
  if (RFCNB_Timeout > 0)
     alarm (RFCNB_Timeout);

  while (1)
  {
    memset (&hdr, 0, sizeof(hdr));
    read_len = read_s (con->fd, (void*)&hdr, sizeof(hdr));
    if (read_len <= 0)  /* Closed of EOF */
       goto bad_read;

    if (hdr.type != RFCNB_SESSION_KEEP_ALIVE)
       break;
    SMB_DEBUG (2, ("RFCNB KEEP ALIVE received\n"));
  }

  /* Got less than we asked for.
   */
  if (read_len < sizeof(hdr))
  {
    /* Now we need to copy the hdr portion we got into the supplied packet
     */
    SMB_ASSERT (pkt->len >= sizeof(hdr));

    memset (pkt->data, 0, sizeof(hdr));
    memcpy (pkt->data, &hdr, read_len);

    RFCNB_Print_Pkt (stdout, "received", pkt, read_len);
    rc = read_len;
    goto quit;
  }

  /* Now, if we got at least a hdr size, alloc space for rest,
   * if we need it
   */
  pkt_len = ntohs (hdr.len);

  SMB_DEBUG (2, ("Reading Pkt: length = %d\n", pkt_len));

  /* Now copy in the header
   */

  SMB_ASSERT (pkt->len >= sizeof(hdr));
  memcpy (pkt->data, &hdr, sizeof(hdr));

  /* Get the rest of the packet ... first figure out how big our buf is?
   * And make sure that we handle the fragments properly.
   */

  if (len < pkt_len)            /* Only get as much as we have space for */
       more = len - sizeof(hdr);
  else more = pkt_len;

  this_time = 0;

  /* We read for each fragment ...
   */
  if (pkt->len == read_len)
  {                             /* If this frag was exact size */
    pkt_frag = pkt->next;       /* Stick next lot in next frag */
    offset = 0;                 /* then we start at 0 in next  */
  }
  else
  {
    pkt_frag = pkt;             /* Otherwise use rest of this frag */
    offset = RFCNB_Pkt_Hdr_Len; /* Otherwise skip the header */
  }

  frag_len = pkt_frag->len;

  if (more <= frag_len)         /* If len left to get less than frag space */
       this_len = more;         /* Get the rest ... */
  else this_len = frag_len - offset;

  while (more > 0)
  {
    SMB_ASSERT (pkt_frag && pkt_frag->len >= offset+this_len);

    this_time = read_s (con->fd, (char*)(pkt_frag->data + offset), this_len);
    if (this_time <= 0)
    {
      if (errno == EINTR || errno == ETIMEDOUT)
         RFCNB_errno = RFCNBE_Timeout;
      else
      {
        if (this_time < 0)
             RFCNB_errno = RFCNBE_BadRead;
        else RFCNB_errno = RFCNBE_ConGone;
      }
      RFCNB_saved_errno = errno;
      rc = -1;
      goto quit;
    }

    SMB_DEBUG (2, ("frag_len = %d, this_time = %d, this_len = %d, more = %d\n",
                   frag_len, this_time, this_len, more));

    read_len += this_time;  /* How much have we read ... */

    if (!pkt_frag->next)    /* stop if no next part */
       break;

    pkt_frag = pkt_frag->next;
    this_len = pkt_frag->len;
    more    -= this_time;
    offset   = 0;
  }

  SMB_DEBUG (2, ("pkt_len = %d, read_len = %d\n", pkt_len, read_len));
  RFCNB_Print_Pkt (stdout, "received", pkt, read_len + sizeof(hdr));

  if (read_len < pkt_len + sizeof(hdr))  /* Discard the rest */
       rc = DiscardRest (con, pkt_len + sizeof(hdr) - read_len);
  else rc = read_len + sizeof(hdr);
  goto quit;

bad_read:
  SMB_DEBUG (1, ("read_s %d bytes: %s\n", read_len, strerror(errno)));

  if (errno == EINTR || errno == ETIMEDOUT)
       RFCNB_errno = RFCNBE_Timeout;
  else RFCNB_errno = RFCNBE_BadRead;
  RFCNB_saved_errno = errno;
  rc = -1;

quit:
  if (RFCNB_Timeout > 0)
     alarm (0);                  /* Reset that sucker */

  return (rc);
}


/*
 * Set up a session with a remote name. We are passed 'called_name' as a
 * string which we convert to a NetBIOS name, i.e. space terminated, up to
 * 16 characters only if we need to. If 'called_addr' is not empty, then
 * we use it to connect to the remote end, but put in 'called_name'.
 * 'called_addr' can be a DNS based name, or a TCP/IP address.
 */
struct RFCNB_Con *RFCNB_Call (const char *called_name,
                              const char *calling_name,
                              const char *called_addr,
                              unsigned short port)
{
  struct RFCNB_Con     *con;
  struct redirect_addr *redir_addr;
  struct in_addr        dest_IP;
  int    sock;
  const  char *service_addr;

  if (port == 0)
     port = SMBlib_def_port;

  /* Create a connection structure first
   */
  con = calloc (sizeof(*con), 1);
  if (!con)
  {
    RFCNB_errno = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno;
    return (NULL);
  }

  /* Resolve that name into an IP address
   */
  service_addr = called_name;
  if (called_addr && *called_addr)
     service_addr = called_addr;

  if (RFCNB_Name_To_IP(service_addr, &dest_IP) < 0)
  {
    RFCNB_saved_errno = errno;
    free (con);
    return (NULL);
  }

  /* Now try to connect to the remote.
   * Connect and get session info etc
   */
  while (1)
  {
    BOOL redirect = FALSE;           /* Assume all OK */

    /* Build the redirect info. First one is first addr called
     * And tack it onto the list of addresses we called
     */
    redir_addr = calloc (sizeof(*redir_addr), 1);
    if (!redir_addr)
    {
      RFCNB_errno = RFCNBE_NoSpace;
      RFCNB_saved_errno = errno;
      free (con);
      return (NULL);
    }

    redir_addr->ip_addr.s_addr = dest_IP.s_addr;
    redir_addr->port           = port;

    if (!con->redirect_list)    /* Stick on head */
         con->redirect_list   = redir_addr;
    else con->last_addr->next = redir_addr;

    con->last_addr = redir_addr;

    /* Now, make that connection
     */
    sock = RFCNB_IP_Connect (dest_IP, port);
    if (sock < 0)
    {
      free (con);
      free (redir_addr);
      return (NULL);  /* RFCNB_errno already set */
    }

    con->fd = sock;

    /* Now send and handle the RFCNB session request.
     * If we get a redirect, we will comeback with redirect true
     * and a new IP address in dest_IP
     */
    if (RFCNB_Session_Req (con, called_name, calling_name,
                           &redirect, &dest_IP, &port) < 0)
    {
      RFCNB_saved_errno = errno;
      free (con);
      free (redir_addr);
      return (NULL);
    }

    if (!redirect)
       break;

    /* We have to close the connection, and then try again
     */
    con->redirects++;
    RFCNB_Close (con->fd);
  }
  return (con);
}

/*
 * We send a packet to the other end ... for the moment, we treat the
 * data as a series of pointers to blocks of data ... we should check the
 * length ...
 */
int RFCNB_Send (const struct RFCNB_Con *con, const struct RFCNB_Pkt *pkt2, int length)
{
#ifdef USE_STRUCTS
  struct RFCNB_Pkt pkt;
  struct RFCNB_Hdr hdr;
#else
  struct RFCNB_Pkt *pkt;
  char  *hdr;
#endif

  int  len;

  SMB_ASSERT (pkt2);
  SMB_ASSERT (pkt2->len == length);

#ifndef USE_STRUCTS
  pkt = RFCNB_Alloc_Pkt (RFCNB_Pkt_Hdr_Len);
  if (!pkt)
     return (-1);

  pkt->next = (struct RFCNB_Pkt*) pkt2;    /* The user data we want to send */
#else
  memset (&pkt, 0, sizeof(pkt));
  memset (&hdr, 0, sizeof(hdr));
  pkt.data  = (BYTE*) &hdr;
  pkt.len   = sizeof(hdr);
  pkt.next  = (struct RFCNB_Pkt*) pkt2;
  hdr.type  = RFCNB_SESSION_MESSAGE;
  hdr.flags = 0;
  hdr.len   = length;
#endif

#ifndef USE_STRUCTS
  hdr = (char*) pkt->data;

  /* Following crap is for portability across multiple UNIX machines
   */
  *(hdr + RFCNB_Pkt_Type_Offset) = RFCNB_SESSION_MESSAGE;
  RFCNB_Put_Pkt_Len (hdr, length);
#endif

  SMB_DEBUG (2, ("pkt2 %p%/%d/%d, Sending packet: ",
                 pkt2->data, pkt2->len, length));

#ifdef USE_STRUCTS
  len = RFCNB_Put_Pkt (con, &pkt, length);
  SMB_ASSERT (len == pkt.len + length);

#else
  len = RFCNB_Put_Pkt (con, pkt, length);
  SMB_ASSERT (len == pkt->len + length);

  pkt->next = NULL;
  RFCNB_Free_Pkt (pkt);
#endif

  return (len);
}

#ifdef NEW_API
int RFCNB_Send2 (const struct RFCNB_Con *con, const void *buf, size_t len)
{
  int tx_len = send (con->fd, buf, len, 0);
  if (tx_len < len)
  {
    if (errno == EINTR || errno == ETIMEDOUT)
         RFCNB_errno = RFCNBE_Timeout;
    else RFCNB_errno = RFCNBE_BadWrite;
    RFCNB_saved_errno = errno;
  }
  return (tx_len);
}

int RFCNB_Recv2 (const struct RFCNB_Con *con, void *buf, size_t max_len)
{
}
#endif

/*
 * We pick up a message from the internet ... We have to worry about
 * non-message packets ...
 */
int RFCNB_Recv (const struct RFCNB_Con *con, struct RFCNB_Pkt *data, int length)
{
  struct RFCNB_Pkt *pkt;
  int    rc;

  if (!con)
  {
    RFCNB_errno = RFCNBE_BadHandle;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  /* Now get a packet from below. We allocate a header first
   * Plug in the header and send the data
   */
  pkt = RFCNB_Alloc_Pkt (RFCNB_Pkt_Hdr_Len);
  if (!pkt)
     return (-1);

  SMB_ASSERT (data->len >= length);

  pkt->next = data;           /* Plug in the data portion */
  rc = RFCNB_Get_Pkt (con, pkt, pkt->len + length);
  pkt->next = NULL;

  if (rc < 0)
  {
    SMB_DEBUG (1, ("Bad packet return in RFCNB_Recv... \n"));
    rc = -1;
  }

  /* We should check that we go a message and not a keep alive
   */
  RFCNB_Free_Pkt (pkt);
  return (rc);
}

/*
 * We just disconnect from the other end, as there is nothing in the RFCNB
 * protocol that specifies any exchange as far as I can see
 */
int RFCNB_Hangup (struct RFCNB_Con *con)
{
  int rc = 0;

  if (con && con->fd > -1)
  {
    rc = RFCNB_Close (con->fd);
    con->fd = -1;
  }
  return (rc);
}

/*
 * Set TCP_NODELAY on the socket
 */
int RFCNB_Set_Sock_NoDelay (const struct RFCNB_Con *con, BOOL yn)
{
  int on = yn;
  return setsockopt (con->fd, IPPROTO_TCP, TCP_NODELAY,
                     (char*)&on, sizeof(on));
}

/*
 * Listen for a connection on a port???, when
 * the connection comes in, we return with the connection
 */
struct RFCNB_Con *RFCNB_Listen (unsigned short port)
{
  ARGSUSED (port);
  return (NULL);
}

/*
 * Pick up the last error response as a string, hmmm, this routine should
 * have been different ...
 */
void RFCNB_Get_Error (char *buffer, int buf_len)
{
  char *p = buffer;

  SMB_ASSERT (RFCNB_errno >= 0);
  SMB_ASSERT (RFCNB_errno < DIM(error_strings));

#ifdef HAVE_SNPRINTF
  p += snprintf (p, buf_len, "%s", error_strings[RFCNB_errno]);
  if (p > buffer && RFCNB_saved_errno > 0)
     snprintf (p, buf_len - (p - buffer),
               "\n\terrno: %s", strerror(RFCNB_saved_errno));
#else
  p += sprintf (p, "%s", error_strings[RFCNB_errno]);
  if (RFCNB_saved_errno > 0)
     sprintf (p, "\n\terrno: %s", strerror(RFCNB_saved_errno));
  ARGSUSED (buf_len);
#endif
}

/*
 * Pick up the last error response and returns as a code
 */
int RFCNB_Get_Last_Error (void)
{
  return (RFCNB_errno);
}

/*
 * Pick up saved errno as well
 */
int RFCNB_Get_Last_Errno (void)
{
  return (RFCNB_saved_errno);
}

/*
 * Pick up the last error response and return in string ...
 */
int RFCNB_Get_Error_Msg (int code, char *msg_buf, int len)
{
  code = abs (code);
  SMB_ASSERT (code < DIM(error_strings));
  strlcpy (msg_buf, error_strings[code], len);
  return (code);
}

/*
 * Register a higher level protocol print routine
 */
void RFCNB_Register_Print_Routine (ProtocolPrinter fn)
{
  Prot_Print_Routine = fn;
}

/*
 * Convert a name and pad to 16 chars ("CA") as needed.
 * name1 is a C string with null termination,
 * name2 holds result of conversion.
 */
static void CvtPadName (const char *name1, char *name2)
{
  char c1, c2;
  int  i, len = strlen (name1);

  for (i = 0; i < 16; i++)
  {
    if (i >= len)
    {
      c1 = 'C';
      c2 = 'A';       /* CA is a space */
    }
    else
    {
      char c = name1[i];
      c1 = (int)(c / 16) + 'A';
      c2 = (int)(c % 16) + 'A';
    }
    name2[i*2]   = c1;
    name2[i*2+1] = c2;
  }
  name2[32] = '\0';
}

/*
 * Converts an ASCII NetBios Name (16 chars) to an RFCNB Name (32 chars)
 * Uses the encoding in RFC1001. Each nibble of byte is added to 'A'
 * to produce the next byte in the name.
 */
void RFCNB_AName_To_NBName (const char *aName, char *nbName, int nblen)
{
  int i;

  SMB_ASSERT (nblen > 32);

  for (i = 0; i < 16; i++)
  {
    char c  = aName[i];
    char c1 = (c >> 4) + 'A';
    char c2 = (c & 0xF) + 'A';

    nbName[i*2]   = c1;
    nbName[i*2+1] = c2;
  }
  nbName[32] = '\0';
}

/*
 * Do the reverse of the above
 */
void RFCNB_NBName_To_AName (const char *nbName, char *aName, int alen)
{
  int i;

  for (i = 0; i < 16 && i < alen-1; i++)
  {
    char c1  = nbName[i*2];
    char c2  = nbName[i*2+1];
    aName[i] = 16 * ((int)c1 - 'A') + ((int)c2 - 'A');
  }
  aName[i] = '\0';
}

/*
 * Print a string of bytes in HEX etc
 */
static void PrintHex (FILE *fd, const struct RFCNB_Pkt *pkt, int ofs, int len)
{
  static char hexChar[] = "0123456789ABCDEF";
  char   c1, c2, outbuf1[33];
  BYTE   c;
  int    i, j = 0;

  SMB_ASSERT (ofs >= 0);

  /* We only want to print as much as specified in len
   */
  while (pkt)
  {
    for (i = 0; i < min(len,pkt->len) - ofs; i++)
    {
      c  = pkt->data [i + ofs];
      c1 = hexChar [c >> 4];
      c2 = hexChar [c & 0xF];

      outbuf1 [j++] = c1;
      outbuf1 [j++] = c2;

      if (j >= sizeof(outbuf1)-1)  /* reached right column */
      {
        fprintf (fd, "    %.*s\n", j, outbuf1);
        j = 0;
      }
    }
    ofs  = 0;
    len -= pkt->len;
    pkt  = pkt->next;
  }

  /* Print last lot in the buffer
   */
  if (j > 0)
     fprintf (fd, "    %.*s\n", j, outbuf1);
  fputc ('\n', fd);
}

static void dump_pkt (const struct RFCNB_Pkt *pkt)
{
  const BYTE *data = (const BYTE*) pkt->data;
  DWORD ofs, datalen = pkt->len + 4;   /* dump marker too */
  DWORD *marker;

  marker = (DWORD*) (pkt->data + pkt->len);
  printf ("dump_pkt(): pkt->len=%d, pkt->data=%p, pkt->next=%p, marker=0x%08lX\n",
          pkt->len, pkt->data, pkt->next, *marker);

  for (ofs = 0; ofs < datalen; ofs += 16)
  {
    size_t j;

    if (ofs == 0)
         printf ("%lu:%s%04lX: ", datalen,
                 datalen > 9999 ? " "    :
                 datalen > 999  ? "  "   :
                 datalen > 99   ? "   "  :
                 datalen > 9    ? "    " :
                                  "     ",  ofs);
    else printf ("       %04lX: ", ofs);

    for (j = 0; j < 16 && j+ofs < datalen; j++)
        printf ("%02X%c", (unsigned)data[j+ofs], j == 7 ? '-' : ' ');  /* no beeps */

    for ( ; j < 16; j++)       /* pad line to 16 positions */
        fputs ("   ", stdout);

    for (j = 0; j < 16 && j+ofs < datalen; j++)
    {
      int ch = data[j+ofs];

      if (ch < ' ')            /* non-printable */
           putc ('.', stdout);
      else putc (ch, stdout);
    }
    putc ('\n', stdout);
  }
}

static void check_pkt (int i, const struct RFCNB_Pkt *pkt)
{
  const DWORD *marker;

  if (!pkt->data)
  {
    fprintf (stdout, "%s(%u): pkt->data == NULL. i: %d. ", __FILE__, __LINE__, i);
    dump_pkt (pkt);
    abort();
  }

  marker = (const DWORD*) (pkt->data + pkt->len);
  if (*marker != DATA_MARKER)
  {
    fprintf (stdout, "%s(%u): *marker != DATA_MARKER. i: %d. ", __FILE__, __LINE__, i);
    dump_pkt (pkt);
    abort();
  }
}

/*
 * Allocate a packet with room for 'data_size' of data.
 */
struct RFCNB_Pkt *RFCNB_Alloc_Pkt (int data_size)
{
  struct RFCNB_Pkt *pkt;
  DWORD *marker;

#if defined(USE_FORTIFY)
  Fortify_ListAllMemory();  /* !! bug hunt */
#endif

  pkt = calloc (sizeof(*pkt) + data_size + sizeof(*marker) + 4, 1);
  if (!pkt)
  {
    SMBlib_errno = SMBlibE_NoSpace;
    RFCNB_errno  = RFCNBE_NoSpace;
    RFCNB_saved_errno = errno; /* ENOMEM */
    return (NULL);
  }

#if defined(USE_FORTIFY)
  Fortify_LabelPointer (pkt, "RFCNB_Alloc_Pkt (head)");
#endif

  SMB_ASSERT (data_size >= 0 && data_size <= USHRT_MAX);

  pkt->next = NULL;
  pkt->len  = data_size;
  pkt->data = (BYTE*)(pkt + 1);
  marker = (DWORD*) ((BYTE*)(pkt + 1) + data_size);
  *marker = DATA_MARKER;
  return (pkt);
}

/*
 * Free up a packet
 */
void RFCNB_Free_Pkt (struct RFCNB_Pkt *pkt)
{
  struct RFCNB_Pkt *next;
  int    i;

  SMB_ASSERT (pkt);

  for (i = 0; pkt; pkt = next, i++)
  {
    check_pkt (i, pkt);
    pkt->data = NULL;
    next = pkt->next;
    free (pkt);
  }

#if defined(USE_FORTIFY)
  Fortify_ListAllMemory();
#endif
}

/*
 * Print an RFCNB packet
 */
void RFCNB_Print_Pkt (FILE *fd, const char *in_out,
                      const struct RFCNB_Pkt *pkt, int len)
{
  char lname[17];
  int  data_len = 40;
  BOOL sending = (strcmp(in_out,"sent") == 0);

  /* We assume that the first fragment is the RFCNB Header
   * We should loop through the fragments printing them out
   */
  fprintf (fd, "RFCNB Pkt %s: ", in_out);

  switch (RFCNB_Pkt_Type (pkt->data))
  {
    case RFCNB_SESSION_MESSAGE:
         fprintf (fd, "SESSION MESSAGE: Length = %i\n", RFCNB_Pkt_Len (pkt->data));

         data_len = RFCNB_Pkt_Len (pkt->data) - RFCNB_Pkt_Hdr_Len;
#if 0
         PrintHex (fd, pkt, RFCNB_Pkt_Hdr_Len, data_len);
#else
         dump_pkt (pkt);
#endif
         if (Prot_Print_Routine)  /* Print the rest of the packet */
           (*Prot_Print_Routine) (fd, !sending, pkt, RFCNB_Pkt_Hdr_Len,
                                  RFCNB_Pkt_Len(pkt->data) - RFCNB_Pkt_Hdr_Len);
         break;

    case RFCNB_SESSION_REQUEST:
         fprintf (fd, "SESSION REQUEST: Length = %i\n", RFCNB_Pkt_Len (pkt->data));
         RFCNB_NBName_To_AName (pkt->data + RFCNB_Pkt_Called_Offset,
                                lname, sizeof(lname));
         fprintf (fd, "  Called Name: %s\n", lname);
         RFCNB_NBName_To_AName (pkt->data + RFCNB_Pkt_Calling_Offset,
                                lname, sizeof(lname));
         fprintf (fd, "  Calling Name: %s\n", lname);
         break;

    case RFCNB_SESSION_ACK:
         fprintf (fd, "RFCNB SESSION ACK: Length = %i\n", RFCNB_Pkt_Len (pkt->data));
         break;

    case RFCNB_SESSION_REJ:
         fprintf (fd, "RFCNB SESSION REJECT: Length = %i\n", RFCNB_Pkt_Len (pkt->data));

         if (RFCNB_Pkt_Len(pkt->data) < 1)
              fprintf (fd, "   Protocol Error, short Reject packet!\n");
         else fprintf (fd, "   Error = %x\n", CVAL (pkt->data, RFCNB_Pkt_Error_Offset));
         break;

    case RFCNB_SESSION_RETARGET:
         fprintf (fd, "RFCNB SESSION RETARGET: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));

         /* !!to-do: print out the IP address etc and the port? */
         break;

    case RFCNB_SESSION_KEEP_ALIVE:
         fprintf (fd, "RFCNB SESSION KEEP ALIVE: Length = %i\n",
                  RFCNB_Pkt_Len (pkt->data));
         break;

    default:
         break;
  }
}

/*
 * Resolve a name into an address
 */
int RFCNB_Name_To_IP (const char *host, struct in_addr *dest_IP)
{
  DWORD addr = inet_addr (host);

  if (addr == INADDR_NONE)  /* not a dotted IP-address */
  {
    /* !!to-do: Look in LMHOSTS-file, then do a WINS lookup, look in
     *          HOSTS-file, and DNS last.
     */
    struct hostent *hp = gethostbyname (host);

    if (!hp)
    {
      RFCNB_errno = RFCNBE_BadName;
      RFCNB_saved_errno = errno;
      return (-1);
    }
    dest_IP->s_addr = *(u_long*)hp->h_addr;
  }
  else
    dest_IP->s_addr = addr;
  return (0);
}

/*
 * Disconnect the TCP connection to the server
 */
int RFCNB_Close (int socket)
{
  return close_s (socket);
}

/*
 * Connect to the server specified in the IP address.
 * Not sure how to handle socket options etc.
 */
int RFCNB_IP_Connect (struct in_addr dest_IP, int port)
{
  struct sockaddr_in addr;
  int    sock = socket (AF_INET, SOCK_STREAM, 0);

  if (sock < 0)
  {
    RFCNB_errno = RFCNBE_BadSocket;
    RFCNB_saved_errno = errno;
    return (-1);
  }

  memset (&addr, 0, sizeof(addr));
  memcpy (&addr.sin_addr, &dest_IP, sizeof(dest_IP));

  addr.sin_port   = htons (port);
  addr.sin_family = AF_INET;

  /* Now connect to the destination. bind() is implicit which selects
   * a free local port for us.
   */
  if (connect (sock, (struct sockaddr*)&addr, sizeof(addr)) < 0)
  {
    RFCNB_errno = RFCNBE_ConnectFailed;
    RFCNB_saved_errno = errno;
    return (-1);
  }
  return (sock);
}

/*
 * Handle the details of establishing the RFCNB session with
 * remote end
 */
int RFCNB_Session_Req (struct RFCNB_Con *con,
                       const  char      *called_name,
                       const  char      *calling_name,
                       BOOL             *redirect,
                       struct in_addr   *dest_IP,
                       unsigned short   *port)
{
  struct RFCNB_Pkt pkt, res_pkt;
//char   resp[16];  /* should be >=9 bytes */
  int    len;

  struct hdr_session {
    struct RFCNB_Hdr        hdr;
    struct RFCNB_SessionPkt session;
  } request;

  struct hdr_resp {
    struct RFCNB_Hdr        hdr;
    struct RFCNB_SessionPkt resp;
  } response;

  *redirect = FALSE;
  memset (&request, 0, sizeof(request));
  memset (&response, 0, sizeof(response));
  memset (&pkt, 0, sizeof(pkt));
  memset (&res_pkt, 0, sizeof(res_pkt));

#ifndef USE_STRUCTS
  {
    char *sess_pkt = &pkt.data;         /* Get pointer to packet proper */

    sess_pkt [RFCNB_Pkt_Type_Offset] = RFCNB_SESSION_REQUEST;
    RFCNB_Put_Pkt_Len (sess_pkt, RFCNB_Pkt_Sess_Len - RFCNB_Pkt_Hdr_Len);
    sess_pkt [RFCNB_Pkt_N1Len_Offset] = ' ';
    sess_pkt [RFCNB_Pkt_N2Len_Offset] = ' ';

    CvtPadName (called_name,  sess_pkt + RFCNB_Pkt_Called_Offset);
    CvtPadName (calling_name, sess_pkt + RFCNB_Pkt_Calling_Offset);
  }

#else
  pkt.data = (BYTE*) &request;
  pkt.len  = sizeof(request);

  request.hdr.type   = RFCNB_SESSION_REQUEST;
  request.hdr.flags  = 0;
  request.hdr.len    = htons (sizeof(request.session));
  request.session.N1 = ' ';
  request.session.N2 = ' ';
  CvtPadName (called_name,  &request.session.called);
  CvtPadName (calling_name, &request.session.calling);
#endif

  /* Now send the packet
   */
  SMB_DEBUG (2, ("Sending packet.\n"));

#ifndef USE_STRUCTS
  len = RFCNB_Put_Pkt (con, pkt, pkt->len);
  RFCNB_Free_Pkt (pkt);
  pkt = NULL;
#else
  len = RFCNB_Put_Pkt (con, &pkt, pkt.len);
#endif

  if (len < 0)
     return (-1);

  SMB_DEBUG (2, ("Getting packet.\n"));

#ifdef USE_STRUCTS
  res_pkt.data = (BYTE*) &response;
  res_pkt.len  = sizeof (response);
#else
  res_pkt.data = resp;
  res_pkt.len  = sizeof (resp);   /* should never get more than one frag */
#endif

  len = RFCNB_Get_Pkt (con, &res_pkt, res_pkt.len);
  if (len < 0)
     return (-1);

  /* Now analyze the packet
   */
#ifdef USE_STRUCTS
  switch (response.hdr.type)
#else
  switch (RFCNB_Pkt_Type (resp))
#endif
  {
    case RFCNB_SESSION_REJ: /* Didn't like us, too bad */

         /* Why did we get rejected ?
          */
#ifdef USE_STRUCTS
         switch (response.resp.N1)
#else
         switch (CVAL (resp, RFCNB_Pkt_Error_Offset))
#endif
         {
           case 0x80:
                RFCNB_errno = RFCNBE_CallRejNLOCN;
                break;
           case 0x81:
                RFCNB_errno = RFCNBE_CallRejNLFCN;
                break;
           case 0x82:
                RFCNB_errno = RFCNBE_CallRejCNNP;
                break;
           case 0x83:
                RFCNB_errno = RFCNBE_CallRejInfRes;
                break;
           case 0x8F:
                RFCNB_errno = RFCNBE_CallRejUnSpec;
                break;
           default:
                RFCNB_errno = RFCNBE_ProtErr;
                break;
         }
         return (-1);

    case RFCNB_SESSION_ACK: /* Got what we wanted */
         return (0);

    case RFCNB_SESSION_RETARGET: /* Go elsewhere */
         *redirect = TRUE;
#ifdef USE_STRUCTS
         dest_IP->s_addr = *(DWORD*) &response.resp.N1;
         *port = SVAL (&response.resp, RFCNB_Pkt_Port_Offset);
#else
         dest_IP->s_addr = *(DWORD*) &resp[RFCNB_Pkt_IP_Offset];
         *port           = SVAL (resp, RFCNB_Pkt_Port_Offset);
#endif
         return (0);

    default:                     /* A protocol error */
         RFCNB_errno = RFCNBE_ProtErr;
         return (-1);
  }
  /* NOT REACHED */
}

/*
 * Print out an SMB pkt in all its gory detail
 */
void SMB_Print_Pkt (FILE *file, const char *in_out,
                    const struct RFCNB_Pkt *pkt, int len)
{
  /* Well, just how do we do this... print it I suppose
   * 1. Print out the SMB header
   * 2. Print the command
   * 3. Print the other bits in the header
   *
   * etc.
   * steal parser from tcpdump (parse_nbt*.c)
   */
  ARGSUSED (file);
  ARGSUSED (in_out);  /* "sent" / "recv" */
  ARGSUSED (pkt);
  ARGSUSED (len);
}

/*
 * Convert a DOS date/time to a local host type date time for printing
 */
const char *SMB_DOSTimToStr (DWORD dos_time)
{
  static char buf[48];
  static char months[] = "???JanFebMarAprMayJunJulAugSepOctNovDec";

  int sec   =  (dos_time & 0x0000001FUL) * 2;
  int min   =  (dos_time & 0x000007E0UL) >> 5;
  int hour  = ((dos_time & 0x0000F800UL) >> 11);

  int day   =  (dos_time & 0x001F0000UL) >> 16;
  int month =  (dos_time & 0x01E00000UL) >> 21;
  int year  = ((dos_time & 0xFE000000UL) >> 25);

  year += (year < 50 ? 2000 : 1980);            /* Y2K hack! */
  if (month > 12)
      month = 0;

  sprintf (buf, "%.3s %2d %02d:%02d:%02d %4d",  /* ctime() format */
           months + 3*month, day, hour, min, sec, year);
  return (buf);
}

/*
 * Convert an attribute byte/word etc to a string. We return a pointer
 * to a static string which we guarantee is long enough. If verbose is
 * true, we print out long form of strings.
 */
const char *SMB_AttrToStr (int attribs, BOOL verbose)
{
  static char buf[128];

  buf[0] = '\0';

  if (attribs & SMB_FA_ROF)
     strcat (buf, (verbose ? "Read Only " : "R"));

  if (attribs & SMB_FA_HID)
     strcat (buf, (verbose ? "Hidden " : "H"));

  if (attribs & SMB_FA_SYS)
     strcat (buf, (verbose ? "System " : "S"));

  if (attribs & SMB_FA_VOL)
     strcat (buf, (verbose ? "Volume " : "V"));

  if (attribs & SMB_FA_DIR)
     strcat (buf, (verbose ? "Directory " : "D"));

  if (attribs & SMB_FA_ARC)
     strcat (buf, (verbose ? "Archive " : "A"));

  return (buf);
}

/*
 * Pick up the Max Buffer Size from the Tree Structure
 */
int SMB_Get_Tree_MBS (const struct SMB_Tree *tree)
{
  if (tree)
     return (tree->mbs);
  return (-1);
}

/*
 * Pick up the Max buffer size
 */
int SMB_Get_Max_Buf_Siz (const struct SMB_Connect *handle)
{
  if (handle)
     return (handle->max_xmit);
  return (-1);
}

/*
 * Pickup the protocol index from the connection structure
 */
int SMB_Get_Protocol_IDX (const struct SMB_Connect *handle)
{
  if (handle)
     return (handle->prot_IDX);
  return (-1);
}

/*
 * Pick up the protocol from the connection structure
 */
int SMB_Get_Protocol (const struct SMB_Connect *handle)
{
  if (handle)
     return (handle->protocol);
  return (SMB_P_Unknown);
}

const char *SMB_Get_ProtocolStr (int proto)
{
  static char buf[20];

  if (proto >= 0 && proto < DIM(SMB_protocols))
     return (SMB_protocols[proto].name);

  sprintf (buf, "proto %d??", proto);
  return (buf);
}

/*
 * Send a TCON to the remote server
 */
struct SMB_Tree *SMB_TreeConnect (struct SMB_Connect *handle,
                                  struct SMB_Tree    *tree,
                                  const  char        *path,
                                  const  char        *password,
                                  const  char        *device)
{
  struct RFCNB_Pkt *pkt;
  struct SMB_Tree  *tree0;
  int    param_len, pkt_len;
  char  *p;

  /* Figure out how much space is needed for path, password, dev
   */
  if (!path || !password || !device)
  {
    SMB_DEBUG (1, ("Bad parameter passed to SMB_TreeConnect\n"));
    SMBlib_errno = SMBlibE_BadParam;
    return (NULL);
  }

  /* The + 2 is because of the \0 and the marker
   */
  param_len = strlen (path) + 2 +
              strlen (password) + 2 +
              strlen (device) + 2;

  /* The -1 accounts for the one byte smb_buf we have because some systems
   * don't like char msg_buf[]
   */
  pkt_len = SMB_tcon_len + param_len;
  RFCNB_ALLOC_PKT (pkt, pkt_len, return (NULL));

  /* Now allocate a tree for this to go into
   */
  if (!tree)
  {
    tree0 = calloc (sizeof(*tree), 1);
    if (!tree0)
    {
      RFCNB_FREE_PKT (pkt);
      SMBlib_errno = SMBlibE_NoSpace;
      return (NULL);
    }
  }
  else
  {
    tree0 = tree;
    tree->next = tree->prev = NULL;
  }

  tree0->con = handle;
  strlcpy (tree0->path,        path,   sizeof(tree0->path));
  strlcpy (tree0->device_type, device, sizeof(tree0->device_type));

  /* Now plug in the values
   */
  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBtcon;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 0;

  SSVAL (SMB_Hdr (pkt), SMB_tcon_bcc_offset, param_len);

  /* Now copy the param strings in with the right stuff
   */
  p = SMB_Hdr (pkt) + SMB_tcon_buf_offset;

  *p++ = SMBasciiID;
  strcpy (p, path);
  p += strlen (path) + 1;

  *p++ = SMBasciiID;
  strcpy (p, password);
  p += strlen (password) + 1;

  *p++ = SMBasciiID;
  strcpy (p, device);

  /* Now send the packet and sit back
   */
  if (RFCNB_Send (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error sending TCon request\n"));

    if (!tree)
       free (tree0);
    RFCNB_FREE_PKT (pkt);
    SMBlib_errno = -SMBlibE_SendFailed;
    return (NULL);
  }

  /* Now get the response
   */
  if (RFCNB_Recv (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error receiving response to TCon\n"));

    if (!tree)
       free (tree0);
    RFCNB_FREE_PKT (pkt);
    SMBlib_errno = -SMBlibE_RecvFailed;
    return (NULL);
  }

  /* Check out the response type
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_DEBUG (1, ("SMB_TCon fail: errorclass = %i, Error Code = %i\n",
                   CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
                   SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset)));

    if (!tree)
       free (tree0);
    SMBlib_SMB_Error = IVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset);
    RFCNB_FREE_PKT (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (NULL);
  }

  tree0->tid = SVAL (SMB_Hdr(pkt), SMB_tconr_tid_offset);
  tree0->mbs = SVAL (SMB_Hdr(pkt), SMB_tconr_mbs_offset);

  SMB_DEBUG (2, ("TConn succeeded, with TID=%i, Max Xmit=%ld\n",
                  tree0->tid, tree0->mbs));

  /* Now link the Tree to the Server Structure
   */
  if (!handle->first_tree)
  {
    handle->first_tree = tree0;
    handle->last_tree  = tree0;
  }
  else
  {
    handle->last_tree->next = tree0;
    tree0->prev = handle->last_tree;
    handle->last_tree = tree0;
  }

  RFCNB_FREE_PKT (pkt);
  return (tree0);
}

int SMB_TreeDisconnect (struct SMB_Tree *tree, BOOL discard)
{
  struct RFCNB_Pkt *pkt;
  int    pkt_len = SMB_tdis_len;

  RFCNB_ALLOC_PKT (pkt, pkt_len, return (-1));

  /* Now plug in the values
   */
  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBtdis;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, tree->con->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, tree->con->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, tree->con->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 0;

  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, tree->tid);
  SSVAL (SMB_Hdr (pkt), SMB_tcon_bcc_offset, 0);

  /* Now send the packet and sit back
   */
  if (RFCNB_Send (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error sending TDis request\n"));

    RFCNB_FREE_PKT (pkt);
    SMBlib_errno = -SMBlibE_SendFailed;
    return (-1);
  }

  /* Now get the response
   */
  if (RFCNB_Recv (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error receiving response to TCon\n"));

    RFCNB_FREE_PKT (pkt);
    SMBlib_errno = -SMBlibE_RecvFailed;
    return (-1);
  }

  /* Check out the response type
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_DEBUG (1, ("SMB_TDis fail: errorclass = %i, Error Code = %i\n",
                   CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
                   SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset)));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    RFCNB_FREE_PKT (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (-1);
  }

  tree->tid = 0xFFFF;    /* Invalidate TID */
  tree->mbs = 0;         /* Invalidate     */

  SMB_DEBUG (2, ("Tree disconnect successful...\n"));

  if (discard)
  {
    /* Unlink it and free tree handle
     */
    if (!tree->next)
         tree->con->first_tree = tree->prev;
    else tree->next->prev      = tree->prev;

    if (!tree->prev)
         tree->con->last_tree = tree->next;
    else tree->prev->next     = tree->next;
  }
  RFCNB_FREE_PKT (pkt);
  return (0);
}


