/*
 * SMBlib NetBIOS implementation
 *
 * Version 1.0
 * SMBlib Routines
 *
 * Copyright (C) Richard Sharpe 1996
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"

#define SMB_LMAPI_SLOT           "\\PIPE\\LANMAN"
#define SMB_LMAPI_SUPW_DESC      "zb16b16WW"

#define SMB_LMAPI_SUI_DESC       "zWsTPWW"
#define SMB_LMAPI_SUI_DATA_DESC  "B16"

int SMBlib_errno     = 0;
int SMBlib_SMB_Error = 0;
int SMBlib_debug     = 0;
int SMBlib_def_port  = 139;  /* netbios-ssn */
int SMB_node_type    = 0;    /* B-node */

static char *SMB_def_domain = SMBLIB_DEFAULT_DOMAIN;
static char *SMB_def_server = NULL;
static char *SMB_def_share  = NULL;

static char *SMB_def_myhost = NULL;
static char *SMB_def_user   = NULL;
static char *SMB_def_passwd = NULL;

static char *SMB_def_osname = SMBLIB_DEFAULT_OSNAME;
static char *SMB_def_lmType = SMBLIB_DEFAULT_LMTYPE;

static FILE *SMB_debug_fout;

static int FigureProtocol (const char **table, int prot_index);

/*
 * WATTCP.CFG config hook handler. Parses "SMB.xx" keywords
 */
static void (*prev_cfg_hook) (const char*, const char*) = NULL;
static void (*prev_post_hook) (void)                    = NULL;

static void smb_config (const char *name, const char *value)
{
  static const struct config_table smb_cfg[] = {
               { "DOMAIN",   ARG_STRDUP,  (void*)&SMB_def_domain     },
               { "CALLED",   ARG_STRDUP,  (void*)&SMB_def_server     },
               { "CALLING",  ARG_STRDUP,  (void*)&SMB_def_myhost     },
               { "SHARE",    ARG_STRDUP,  (void*)&SMB_def_share      },
               { "USERNAME", ARG_STRDUP,  (void*)&SMB_def_user       },
               { "PASSWORD", ARG_STRDUP,  (void*)&SMB_def_passwd     },
               { "OSNAME",   ARG_STRDUP,  (void*)&SMB_def_osname     },
               { "LMTYPE",   ARG_STRDUP,  (void*)&SMB_def_lmType     },
               { "LMHOSTS",  ARG_FUNC,    (void*)SMB_ReadLMhosts     },
               { "PORT",     ARG_ATOI,    (void*)&SMBlib_def_port    },
               { "TIMEOUT",  ARG_ATOI,    (void*)&RFCNB_Timeout      },
               { "NODETYPE", ARG_ATOI,    (void*)&SMB_node_type      },
               { "DEBUG",    ARG_ATOI,    (void*)&SMBlib_debug       },
               { NULL }
             };

  if (!parse_config_table(smb_cfg, "SMB.", name, value) &&
      prev_cfg_hook)
     (*prev_cfg_hook) (name, value);
}

static void smb_post_init (void)
{
  if (!SMB_def_myhost)
  {
    fprintf (stdout, "SMB-lib warning: \"SMB.CALLING\" not specified. "
             "Must use DNS hostname as calling name.\n");
  }
}

/*
 * Initialize the SMBlib package
 */
W32_DATA int _watt_is_init;

int SMB_Init (void)
{
  SMB_debug_fout = stdout;

  if (_watt_is_init)
  {
    fprintf (stdout, "SMB_Init() must be called before sock_init()\n");
    return (-1);
  }

  prev_cfg_hook  = usr_init;
  prev_post_hook = usr_post_init;
  usr_init       = smb_config;
  usr_post_init  = smb_post_init;

  sock_init();

#if defined(SMB_WIN32) && defined(_MSC_VER)
  LoadLibrary ("exc-abort.dll");
#endif
  return (0);
}

/*
 * Print some debug
 */
void SMB_debug (const char *fmt, ...)
{
  va_list arg;
  va_start (arg, fmt);
  vfprintf (SMB_debug_fout, fmt, arg);
  va_end (arg);
}

/*
 * Open and parse the "SMB.LMHOSTS = file" entry in config-file.
 */
void SMB_ReadLMhosts (const char *fname)
{
  /* !! todo */
}

/*
 * SMB_Create: Create a connection structure and return for later use
 * We have other helper routines to set variables
 */
struct SMB_Connect *SMB_Create_Con_Handle (void)
{
  SMBlib_errno = SMBlibE_NotImpl;
  return (NULL);
}

int SMBlib_Set_Sock_NoDelay (struct SMB_Connect *con, BOOL yn)
{
  if (RFCNB_Set_Sock_NoDelay (con->transConnect, yn) < 0)
     SMB_DEBUG (1, ("Setting no-delay on TCP socket fail: %s",
                strerror(errno)));
  return (0);
}

/*
 * Get our hostname. Not the FQDN, but NetBIOS name.
 */
int SMB_Get_My_Name (char *name, int len)
{
  W32_DATA char _w32_hostname[80+1];  /* in libwatt.a, pctcp.c */

  if (SMB_def_myhost)
       strlcpy (name, SMB_def_myhost, len);
  else strlcpy (name, _w32_hostname, len);
  return strlen (name);
}

/*
 * SMB_Connect_Server: Connect to a server, but don't negotiate
 * protocol or anything else ...
 */
struct SMB_Connect *SMB_Connect_Server (struct SMB_Connect *handle,
                                        const  char        *server)
{
  char called[80], calling[80], *address;

  /* Get a connection structure if one does not exist
   */
  if (!handle)
  {
    handle = calloc (sizeof(*handle), 1);
    if (!handle)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (NULL);
    }
  }

  /* Init some things ...
   */
  handle->service [0] = '\0';
  handle->username[0] = '\0';
  handle->password[0] = '\0';
  handle->address [0] = '\0';
  handle->sock_options[0] = '\0';

  strlcpy (handle->desthost, server, sizeof(handle->desthost)); /* DNS name or IP */
  strlcpy (handle->PDomain, SMB_def_domain, sizeof(handle->PDomain));
  strlcpy (handle->OSName,  SMB_def_osname, sizeof(handle->OSName));
  strlcpy (handle->LMType,  SMB_def_lmType, sizeof(handle->LMType));
  handle->first_tree = handle->last_tree = NULL;

  SMB_Get_My_Name (handle->myname, sizeof(handle->myname));

  handle->port = 0;                /* No port selected */

  /* Get some things we need for the SMB Header
   */
  handle->pid = getpid();
  handle->mid = handle->pid;
  handle->uid = 0;
#ifdef __DJGPP__
  handle->gid = getgid();
#endif

  /* Now connect to the remote end, but first upper case the name of the
   * service we are going to call, since some servers want it in uppercase
   */
#if 0
  /* map IP/DNS name to NetBIOS name
   */
  if (!SMB_Get_His_Name (server, called, sizeof(called)))
  {
    free (handle);
    SBlib_errno = ??
    return (NULL);
  }
#else
  if (!SMB_def_server || !server)
  {
    free (handle);
    SMBlib_errno = -SMBlibE_CallFailed;
    return (NULL);
  }
  strlcpy (called, SMB_def_server, sizeof(called));
  strupr (called);
#endif

  strlcpy (calling, handle->myname, sizeof(calling));
  strupr (calling);

  address = (handle->address[0] ? handle->address : handle->desthost);

  handle->transConnect = RFCNB_Call (called, calling, address, 0);
  if (!handle->transConnect)
  {
    free (handle);
    SMBlib_errno = -SMBlibE_CallFailed;
    return (NULL);
  }
  return (handle);
}

/*
 * SMB_Connect: Connect to the indicated server
 * If Con_Handle == NULL then create a handle and connect, otherwise
 * use the handle passed
 */
static const char *prots_restrict[] = { "PC NETWORK PROGRAM 1.0",
                                        NULL
                                      };

struct SMB_Connect *SMB_Connect (struct SMB_Connect *handle,
                                 struct SMB_Tree   **tree,
                                 const  char        *service,
                                 const  char        *username,
                                 const  char        *password)
{
  char *host, *address;
  char  temp[80], called[80], calling[80];
  int   i;

  /* Get a connection structure if one does not exist
   */
  if (!handle)
  {
    handle = calloc (sizeof(*handle), 1);
    if (!handle)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (NULL);
    }
  }

  /* Init some things ...
   */
  strlcpy (handle->service,  service, sizeof(handle->service));
  strlcpy (handle->username, username, sizeof(handle->username));
  strlcpy (handle->password, password, sizeof(handle->password));
  handle->sock_options[0] = handle->address[0] = '\0';

  strlcpy (handle->PDomain,SMB_def_domain, sizeof(handle->PDomain));
  strlcpy (handle->OSName, SMB_def_osname, sizeof(handle->OSName));
  strlcpy (handle->LMType, SMB_def_lmType, sizeof(handle->LMType));
  handle->first_tree = handle->last_tree = NULL;

  SMB_Get_My_Name (handle->myname, sizeof(handle->myname));

  handle->port = 0;                /* No port selected */

  /* Get some things we need for the SMB Header
   */
  handle->pid = getpid();
  handle->mid = handle->pid;       /* This will do for now ... */
  handle->uid = 0;                 /* Until we have done a logon, no uid */
#ifdef __DJGPP__
  handle->gid = getgid();
#endif

  /* Now figure out the host portion of the service
   */
  strlcpy (temp, service, sizeof(temp)-1);
  host = strtok (temp, "/\\");  /* Separate host name portion */

  SMB_ASSERT (host);
  strlcpy (handle->desthost, host, sizeof(handle->desthost));

  /* Now connect to the remote end, but first upper case the name of the
   * service we are going to call, sine some servers want it in uppercase
   */
  for (i = 0; i < strlen(host); i++)
      called[i] = toupper (host[i]);

  called [i] = '\0';

  for (i = 0; i < strlen(handle->myname); i++)
      calling[i] = toupper (handle->myname[i]);

  calling [i] = '\0';

  if (handle->address[0] == '\0')
       address = handle->desthost;
  else address = handle->address;

  handle->transConnect = RFCNB_Call (called, calling, address, handle->port);
  if (!handle->transConnect)
  {
    SMBlib_errno = -SMBlibE_CallFailed;
    goto fail;
  }

  /* Now, negotiate the protocol
   */
  if (SMB_Negotiate (handle, prots_restrict) < 0)
  {
    /* Hmmm what should we do here ... We have a connection, but
     * could not negotiate ...
     */
     goto fail;
  }

  /* Now connect to the "A:" service ...
   */
  *tree = SMB_TreeConnect (handle, NULL, service, password, "A:");
  if (*tree)
     return (handle);

fail:
  free (handle);
  return (NULL);
}


/***********************************************************************************************/

struct SMB_header {
       struct RFCNB_Hdr nb_hdr;
       BYTE  idf [4];         /* "\xFFSMB" */
       BYTE  command;         /* SMB_COM_x */
       DWORD status;
       BYTE  flags;
       WORD  flags2;
       WORD  pid_high;
       WORD  signature[4];
       WORD  unused;
       WORD  tid;             /* ofs 24 */
       WORD  pid;             /* ofs 26 */
       WORD  uid;             /* ofs 28 */
       WORD  mid;             /* ofs 30 */
    };

static const void *create_SMB_header (BYTE command, const struct SMB_Connect *handle)
{
  static struct SMB_header hdr;

  memset (&hdr, 0, sizeof(hdr));
  hdr.nb_hdr.type = RFCNB_SESSION_MESSAGE;
  hdr.idf[0] = '\xFF';
  hdr.idf[1] = 'S';
  hdr.idf[2] = 'M';
  hdr.idf[3] = 'B';
  hdr.command = command;
  hdr.tid = 0;
  hdr.pid = handle->pid;
  hdr.uid = handle->uid;
  hdr.mid = handle->mid;
  return (const void*) &hdr;
}

/*
 * Negotiate the protocol we will use from the list passed in prots.
 * We return the index of the accepted protocol in NegProt, -1 indicates
 * none acceptible, and our return value is 0 if ok, -1 if problems.
 */
int SMB_Negotiate (struct SMB_Connect *handle, const char **prots)
{
#ifdef USE_STRUCTS
  struct RFCNB_Pkt pkt;

  struct SMB_negotiate_req {
    struct SMB_header hdr;
    BYTE   wct;             /* ofs 32 */
    WORD   bcc;             /* ofs 33 */
    BYTE   buffer[200];     /* ofs 35 */
  } request;

  struct negotiate_resp {
    struct RFCNB_Hdr hdr;            /* 4 bytes */
    WORD             error_class;    /* ofs 5 */
    WORD             error_code;     /* ofs 7 */
    BYTE             unused1[23];    /* ofs 9 */
    BYTE             wct;            /* ofs 32 */
    WORD             CP_idx;         /* ofs 33 */
    WORD             sec_mode;       /* ofs 35 */
    WORD             max_buf_size;   /* ofs 37 */
    WORD             max_mpx_cnt;    /* ofs 39 */
    WORD             max_VCs;        /* ofs 41 */
    WORD             raw_mode;       /* ofs 43 */
    DWORD            sess_key;       /* ofs 45 */
    DWORD            unused2;        /* ofs 49 */
    WORD             time_zone;      /* ofs 53 */
    WORD             enc_key_len;    /* ofs 55 */
    BYTE             unused3[4];     /* ofs 57 */
    BYTE             lm_buf[100];    /* ofs 61 */
  } response;

#else
  struct RFCNB_Pkt *pkt;
#endif

  int   prots_len, i, pkt_len, prot, wct, rc, len;
  char *p, *p_max;
  WORD  cp_idx;

  SMB_ASSERT (prots);

  /* Figure out how long the prot list will be and allocate space for it
   */
  for (i = prots_len = 0; prots[i]; i++)
      prots_len += strlen (prots[i]) + 2;  /* termination + SMBdialectID */

#ifdef USE_STRUCTS
  pkt_len = offsetof (struct SMB_negotiate_req, buffer) + prots_len;
  memset (&pkt, 0, sizeof(pkt));
  memset (&request, 0, sizeof(request));
  pkt.data = (BYTE*) &request;
  pkt.len  = offsetof(struct SMB_negotiate_req, buffer) + prots_len;
#else
  pkt_len = SMB_negp_len + prots_len;
  pkt = RFCNB_Alloc_Pkt (pkt_len);
  if (!pkt)
     return (-1);
#endif

  /* Now plug in the bits we need
   */
#ifdef USE_STRUCTS
  memcpy (&request.hdr, create_SMB_header(SMB_COM_NEGOTIATE,handle), sizeof(request.hdr));
  request.wct = 0;
  request.bcc = htons (prots_len);  // !! added swap

  p = request.buffer;
  p_max = p + sizeof(request.buffer);
#else
  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  SCVAL (SMB_Hdr (pkt), SMB_hdr_com_offset, SMB_COM_NEGOTIATE);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
  SCVAL (SMB_Hdr (pkt), SMB_hdr_wct_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_negp_bcc_offset, prots_len);

  /* Now copy the prot strings in with the right stuff
   */
  p = SMB_Hdr(pkt) + SMB_negp_buf_offset;
  p_max = p + pkt_len;
#endif

  for (i = 0; prots[i]; i++)
  {
    *p++ = SMBdialectID;
    strcpy (p, prots[i]);
    SMB_DEBUG (4, ("Trying to negiotiate \"%.30s\"\n", p));
    p += strlen (p) + 1;
  }

  SMB_ASSERT (p < p_max);

  /* Now send the packet and sit back
   */
#ifdef NEW_API
  len = sizeof(request) - sizeof(request.buffer) + (p - request.buffer);
  request.hdr.nb_hdr.len = htons (len - sizeof(request.hdr.nb_hdr));
  len = RFCNB_Send2 (handle->transConnect, &request, len);
#elif defined(USE_STRUCTS)
  len = RFCNB_Send (handle->transConnect, &pkt, pkt_len);
#else
  len = RFCNB_Send (handle->transConnect, pkt, pkt_len);
#endif

  if (len < 0)
  {
    SMB_DEBUG (1, ("Error sending negotiate protocol\n"));
    SMBlib_errno = -SMBlibE_SendFailed;      /* Failed, check errno */
    rc = -1;
    goto quit;
  }

#ifndef USE_STRUCTS
  RFCNB_Free_Pkt (pkt);
#endif

  /* Now get the response (reuse the send pkt for recv).
   * Make sure that the pkt len is long enough for the max response.
   * Which is a problem, because the encryption key len EEC may be long
   */
  pkt_len = SMB_hdr_wct_offset + (19 * 2) + 40;

#ifndef USE_STRUCTS
  pkt = RFCNB_Alloc_Pkt (pkt_len);
  if (!pkt)
  {
    rc = -1;
    goto quit;
  }
#endif

#ifdef USE_STRUCTS
  SMB_ASSERT (sizeof(response) >= pkt_len);
  pkt.data = (BYTE*) &response;
  pkt.len  = sizeof(response);

  if (RFCNB_Recv (handle->transConnect, &pkt, pkt_len) < 0)
#else
  if (RFCNB_Recv (handle->transConnect, pkt, pkt_len) < 0)
#endif
  {
    SMB_DEBUG (1, ("Error receiving response to negotiate\n"));
    SMBlib_errno = -SMBlibE_RecvFailed;  /* Failed, check errno */
    rc = -1;
    goto quit;
  }

#ifdef USE_STRUCTS
  if (response.error_class != SMBC_SUCCESS)
#else
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
#endif
  {
    int err_code, err_class;
#ifdef USE_STRUCTS
    err_class = response.error_class;
    err_code  = response.error_code;
#else
    err_class = CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset);
    err_code  = SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset);
#endif
    SMB_DEBUG (1, ("fail: errorclass = %d, Error Code = %d\n", err_class, err_code));
    SMBlib_SMB_Error = err_class + (err_code << 16);
    SMBlib_errno = SMBlibE_Remote;
    rc = -1;
    goto quit;
  }

#ifdef USE_STRUCTS
  cp_idx = response.CP_idx;
#else
  cp_idx = SVAL (SMB_Hdr (pkt), SMB_negrCP_idx_offset);
#endif

  if (cp_idx == 0xFFFF)
  {
    SMB_DEBUG (1, ("None of our protocols was accepted\n"));
    SMBlib_errno = SMBlibE_NegNoProt;
    rc = -1;
    goto quit;
  }

  /* Now, unpack the info from the response, if any and evaluate the proto
   * selected. We must make sure it is one we like
   */

  handle->prot_IDX = prot = cp_idx;
  handle->protocol = FigureProtocol (prots, prot);

  if (handle->protocol == SMB_P_Unknown)
  {
    SMBlib_errno = SMBlibE_ProtUnknown;
    rc = -1;
    goto quit;
  }

#ifdef USE_STRUCTS
  switch (response.wct)
#else
  switch (CVAL (SMB_Hdr(pkt), SMB_hdr_wct_offset))
#endif
  {
    case 0x01:              /* No more info */
         break;

    case 13:                /* Up to and including LanMan 2.1 */
#ifdef USE_STRUCTS
         handle->security = response.sec_mode;
         handle->encrypt_passwords = ((handle->security &
                                      SMB_sec_encrypt_mask) != 0x00);
         handle->security   = (handle->security & SMB_sec_user_mask);
         handle->max_xmit   = response.max_buf_size;
         handle->maxMPX     = response.max_mpx_cnt;
         handle->maxVC      = response.max_VCs;
         handle->rawSupport = response.raw_mode;
         handle->sessionKey = response.sess_key;
         handle->serverTZ   = response.time_zone;
         handle->encrKeyLen = response.enc_key_len;
         p = response.lm_buf + response.enc_key_len;
#else
         handle->security = SVAL (SMB_Hdr (pkt), SMB_negrLM_sec_offset);
         handle->encrypt_passwords = ((handle->security &
                                      SMB_sec_encrypt_mask) != 0x00);

         handle->security   = (handle->security & SMB_sec_user_mask);
         handle->max_xmit   = SVAL (SMB_Hdr (pkt), SMB_negrLM_mbs_offset);
         handle->maxMPX     = SVAL (SMB_Hdr (pkt), SMB_negrLM_mmc_offset);
         handle->maxVC      = SVAL (SMB_Hdr (pkt), SMB_negrLM_mnv_offset);
         handle->rawSupport = SVAL (SMB_Hdr (pkt), SMB_negrLM_rm_offset);
         handle->sessionKey = IVAL (SMB_Hdr (pkt), SMB_negrLM_sk_offset);
         handle->serverTZ   = SVAL (SMB_Hdr (pkt), SMB_negrLM_stz_offset);
         handle->encrKeyLen = SVAL (SMB_Hdr (pkt), SMB_negrLM_ekl_offset);

         p = SMB_Hdr(pkt) + SMB_negrLM_buf_offset + handle->encrKeyLen;
#endif
         strlcpy (handle->serverPDom, p, sizeof(handle->serverPDom));  /* no used!! */
         break;

    case 17:                /* NT LM 0.12 and LN LM 1.0 */
#ifdef USE_STRUCTS
         handle->security = response.sec_mode;
         handle->encrypt_passwords = ((handle->security &
                                      SMB_sec_encrypt_mask) != 0x00);
         handle->security   = (handle->security & SMB_sec_user_mask);
         handle->max_xmit   = response.max_buf_size;
         handle->maxMPX     = response.max_mpx_cnt;
         handle->maxVC      = response.max_VCs;
         handle->maxRaw     = *(DWORD*) ((BYTE*)&response + SMB_negrNTLM_mrs_offset);
         handle->sessionKey = *(DWORD*) ((BYTE*)&response + SMB_negrNTLM_sk_offset);
         handle->serverTZ   = *(DWORD*) ((BYTE*)&response + SMB_negrNTLM_stz_offset);
         handle->encrKeyLen = *((BYTE*)&response + SMB_negrNTLM_ekl_offset);

         p = ((char*)&response + SMB_negrLM_buf_offset + handle->encrKeyLen);
#else
         handle->security = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_sec_offset);
         handle->encrypt_passwords = ((handle->security &
                                      SMB_sec_encrypt_mask) != 0x00);

         handle->security   = (handle->security & SMB_sec_user_mask);
         handle->max_xmit   = IVAL (SMB_Hdr (pkt), SMB_negrNTLM_mbs_offset);
         handle->maxMPX     = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_mmc_offset);
         handle->maxVC      = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_mnv_offset);
         handle->maxRaw     = IVAL (SMB_Hdr (pkt), SMB_negrNTLM_mrs_offset);
         handle->sessionKey = IVAL (SMB_Hdr (pkt), SMB_negrNTLM_sk_offset);
         handle->serverTZ   = SVAL (SMB_Hdr (pkt), SMB_negrNTLM_stz_offset);
         handle->encrKeyLen = CVAL (SMB_Hdr (pkt), SMB_negrNTLM_ekl_offset);

         p = SMB_Hdr (pkt) + SMB_negrLM_buf_offset + handle->encrKeyLen;
#endif
         strlcpy (handle->serverPDom, p, sizeof(handle->serverPDom));  /* no used!! */
         break;

    default:
#ifdef USE_STRUCTS
         wct = response.wct;
#else
         wct = CVAL (SMB_Hdr(pkt), SMB_hdr_wct_offset);
#endif
         SMB_DEBUG (1, ("Unknown NegProt response format, ignored. wct = %d\n", wct));
         break;
  }

  SMB_DEBUG (2, ("Protocol selected is: %d:%.30s\n", prot, prots[prot]));
  rc = 0;

quit:
#ifndef USE_STRUCTS
  RFCNB_Free_Pkt (pkt);
#endif
  return (rc);
}

/*
 * Logon to the server. That is, do a session setup if we can.
 * We do not do Unicode yet!
 */
int SMB_Logon_Server (struct SMB_Connect *handle,
                      const  char        *userName,
                      const  char        *passWd)
{
  struct RFCNB_Pkt *pkt;
  int    param_len, pkt_len, rc;
  char  *p;

  /* First we need a packet etc ... but we need to know what protocol has
   * been negotiated to figure out if we can do it and what SMB format to
   * use ...
   */
  if (handle->protocol < SMB_P_LanMan1)
  {
    SMBlib_errno = SMBlibE_ProtLow;
    return (-1);
  }

  /* Now build the correct structure
   */
  if (handle->protocol < SMB_P_NT1)
  {
    /* We don't handle encrypted passwords ...
     */
    param_len = strlen (userName) + 1 +
                strlen (passWd) + 1 +
                strlen (handle->PDomain) + 1 +
                strlen (handle->OSName) + 1;

    pkt_len = SMB_ssetpLM_len + param_len;
    RFCNB_ALLOC_PKT (pkt, pkt_len, return (-1));

    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    SCVAL (SMB_Hdr (pkt), SMB_hdr_com_offset, SMBsesssetupX);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
    SCVAL (SMB_Hdr (pkt), SMB_hdr_wct_offset, 10);
    SCVAL (SMB_Hdr (pkt), SMB_hdr_axc_offset, 0xFF); /* No extra command */
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, 0);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mmc_offset, 2);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_vcn_offset, handle->pid);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_pwl_offset, strlen (passWd) + 1);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_res_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char*) SMB_Hdr (pkt) + SMB_ssetpLM_buf_offset;

    /* Copy in password, then the rest. Password has a null at end
     */
    strcpy (p, passWd);
    p += strlen (passWd);

    strcpy (p, userName);
    p += strlen (userName) + 1;

    strcpy (p, handle->PDomain);
    p += strlen (handle->PDomain) + 1;

    strcpy (p, handle->OSName);
  }
  else
  {
    /* We don't admit to UNICODE support ...
     */
    param_len = strlen (userName) + 1 + strlen (passWd) +
                strlen (handle->PDomain) + 1 +
                strlen (handle->OSName) + 1 +
                strlen (handle->LMType) + 1;

    pkt_len = SMB_ssetpNTLM_len + param_len;
    RFCNB_ALLOC_PKT (pkt, pkt_len, return (-1));

    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 13;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = 0xFF; /* No extra command */
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, 0);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mmc_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_vcn_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cipl_offset, strlen (passWd));
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cspl_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_res_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cap_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char*) SMB_Hdr (pkt) + SMB_ssetpNTLM_buf_offset;

    /* Copy in password, then the rest. Password has no null at end
     */
    strcpy (p, passWd);
    p += strlen (passWd);

    strcpy (p, userName);
    p += strlen (userName) + 1;

    strcpy (p, handle->PDomain);
    p += strlen (handle->PDomain) + 1;

    strcpy (p, handle->OSName);
    p += strlen (handle->OSName) + 1;

    strcpy (p, handle->LMType);
  }

  /* Now send it and get a response
   */
  if (RFCNB_Send (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error sending SessSetupX request\n"));
    SMBlib_errno = SMBlibE_SendFailed;
    rc = -1;
    goto quit;
  }

  /* Now get the response ...
   */
  if (RFCNB_Recv (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG  (1, ("Error receiving response to SessSetupAndX\n"));
    SMBlib_errno = SMBlibE_RecvFailed;
    rc = -1;
    goto quit;
  }

  /* Check out the response type ...
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_DEBUG (1, ("SMB_SessSetupAndX fail: errorclass = %d, Error Code = %d\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset)));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    SMBlib_errno = SMBlibE_Remote;
    rc = -1;
    goto quit;
  }

  SMB_DEBUG (1, ("SessSetupAndX response. Action = %d\n",
             SVAL (SMB_Hdr(pkt), SMB_ssetpr_act_offset)));

  /* Now pick up the UID for future reference ...
   */
  handle->uid = SVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset);
  rc = 0;

quit:
  RFCNB_FREE_PKT (pkt);
  return (rc);
}

/*
 * Disconnect from the server, and disconnect all tree connects
 */
int SMB_Discon (struct SMB_Connect **handle, BOOL keepHandle)
{
  if (!handle || !*handle)
     return (-1);

  RFCNB_Hangup ((*handle)->transConnect);

  free ((*handle)->transConnect);
  (*handle)->transConnect = NULL;

  if (!keepHandle)
  {
    free (*handle);
    *handle = NULL;
  }
  return (0);
}

/*
 * Send an api request to the "\\server\IPC$" tree, with a "\PIPE\LANMAN"
 * API request to change the user's password
 */
int SMBapi_NetUserPasswordSet (struct SMB_Tree *tree,
                               const  char     *user,
                               const  char     *oldpass,
                               const  char     *newpass,
                               int             *apiStatus)
{
  struct RFCNB_Pkt *pkt;
  int    param_len, pkt_len, rc;
  int    pad_api_name = FALSE;
  char  *p;

  /* Get a packet, we need one with space for a transact plus. The calc
   * below lays it all out as it is, including the empty string after the
   * descriptor and before the username
   */
  param_len = 2 + strlen (SMB_LMAPI_SUPW_DESC) + 1 +
              1 /* for empty string :-) */ +
              strlen (user) + 1 + 16 + 16 + 2 + 2;

  /* We have no setup words, we don't account for them
   */
  pkt_len = SMB_trans_len + 2 + strlen (SMB_LMAPI_SLOT) + 1;

  /* Pad things onto a word boundary ...
   */
  if (pkt_len & 1)
  {
    pkt_len++;
    pad_api_name = TRUE;
  }

  pkt_len = pkt_len + param_len;
  RFCNB_ALLOC_PKT (pkt, pkt_len, return (-1));

  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBtrans;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, tree->con->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, tree->tid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, tree->con->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, tree->con->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 14;

  SSVAL (SMB_Hdr (pkt), SMB_trans_tpc_offset, param_len);
  SSVAL (SMB_Hdr (pkt), SMB_trans_tdc_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_trans_mpc_offset, 4);
  SSVAL (SMB_Hdr (pkt), SMB_trans_mdc_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_trans_msc_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_trans_flg_offset, 0);
  SIVAL (SMB_Hdr (pkt), SMB_trans_tmo_offset, 5000);
  SSVAL (SMB_Hdr (pkt), SMB_trans_pbc_offset, param_len);
  SSVAL (SMB_Hdr (pkt), SMB_trans_pbo_offset, SMB_trans_len + 2 + strlen (SMB_LMAPI_SLOT) + 1);
  SSVAL (SMB_Hdr (pkt), SMB_trans_dbc_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_trans_dbo_offset, 0);

  /* Now put in the bcc and the rest of the info ...
   */
  SSVAL (SMB_Hdr(pkt), SMB_trans_len, param_len + strlen(SMB_LMAPI_SLOT)+1);

  p = (char*) SMB_Hdr (pkt) + SMB_trans_len + 2; /* Skip the BCC and ECT */

  strcpy (p, SMB_LMAPI_SLOT);
  p += strlen (SMB_LMAPI_SLOT) + 1;

  if (pad_api_name == TRUE)     /* Pad if we need to */
     p++;

  SSVAL (p, 0, SMB_LMapi_UserPasswordSet); /* The api call */
  p += 2;

  strcpy (p, SMB_LMAPI_SUPW_DESC); /* Copy in the param desc */

  p += strlen (SMB_LMAPI_SUPW_DESC) + 1;

  *p++ = '\0';                    /* Stick in that null string */

  strlcpy (p, user, 128);
  p += strlen (user) + 1;

  strlcpy (p, oldpass, 16);
  p += 16;

  strlcpy (p, newpass, 16);
  p += 16;

  SSVAL (p, 0, 0);                /* Seems to be zero always? */
  SSVAL (p, 2, strlen(newpass));  /* Length of new password */

  /* Now send the lot and get a response ...
   */
  if (RFCNB_Send (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error sending Trans request\n"));
    SMBlib_errno = SMBlibE_SendFailed;
    rc = -1;
    goto quit;
  }

  /* Now get the response ...
   */
  if (RFCNB_Recv (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error receiving response to Trans request\n"));
    SMBlib_errno = SMBlibE_RecvFailed;
    rc = -1;
    goto quit;
  }

  /* Check out the response type ...
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_DEBUG (1, ("SMB_trans fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset)));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    SMBlib_errno = SMBlibE_Remote;
    rc = -1;
    goto quit;
  }

  /* All ok, pass back the status
   */
  *apiStatus = SVAL (SMB_Hdr(pkt), SVAL (SMB_Hdr(pkt),
               SMB_transr_pbo_offset));
  rc = 0;

quit:
  RFCNB_FREE_PKT (pkt);
  return (rc);
}

/*
 * Set user info ... specifically, password
 */
int SMBapi_NetSetUserInfo (struct SMB_Tree *tree,
                           const  char     *user,
                           const  char     *newpass,
                           int             *apiStatus)
{
  struct RFCNB_Pkt *pkt;
  int    param_len, pkt_len, data_len, rc;
  int    pad_api_name = FALSE;
  int    pad_params   = FALSE;
  char  *p;

  /* Get a packet, we need one with space for a transact plus. The calc
   * below lays it all out as it is, including the empty string after the
   * descriptor and before the username
   */
  param_len = 2 + strlen (SMB_LMAPI_SUI_DESC) +
              1 + strlen (SMB_LMAPI_SUI_DATA_DESC) +
              1 + strlen (user) + 1 + 2 + 2 + 2 + 2;
  data_len = 16;

  /* We have no setup words, so we don't account for them
   */
  pkt_len = SMB_trans_len + 2 + strlen (SMB_LMAPI_SLOT) + 1;

  if (pkt_len & 1)            /* Pad to a WORD boundary */
     pad_api_name = TRUE;

  if (param_len & 1)          /* pad to a WORD boundary */
     pad_params = TRUE;

  pkt_len = pkt_len + param_len + data_len;

  if (pad_api_name == TRUE)
     pkt_len++;

  if (pad_params == TRUE)
     pkt_len++;

  RFCNB_ALLOC_PKT (pkt, pkt_len, return (-1));

  SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
  *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBtrans;
  SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, tree->con->pid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, tree->tid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, tree->con->mid);
  SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, tree->con->uid);
  *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 14;

  SSVAL (SMB_Hdr (pkt), SMB_trans_tpc_offset, param_len);
  SSVAL (SMB_Hdr (pkt), SMB_trans_tdc_offset, data_len);
  SSVAL (SMB_Hdr (pkt), SMB_trans_mpc_offset, 4);
  SSVAL (SMB_Hdr (pkt), SMB_trans_mdc_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_trans_msc_offset, 0);
  SSVAL (SMB_Hdr (pkt), SMB_trans_flg_offset, 0);
  SIVAL (SMB_Hdr (pkt), SMB_trans_tmo_offset, 5000);
  SSVAL (SMB_Hdr (pkt), SMB_trans_pbc_offset, param_len);
  SSVAL (SMB_Hdr (pkt), SMB_trans_pbo_offset, SMB_trans_len + 2 + strlen (SMB_LMAPI_SLOT) + 1);
  SSVAL (SMB_Hdr (pkt), SMB_trans_dbc_offset, data_len);
  SSVAL (SMB_Hdr (pkt), SMB_trans_dbo_offset, pkt_len - data_len);

  /* Now put in the bcc and the rest of the info ...
   */
  SSVAL (SMB_Hdr(pkt), SMB_trans_len,
         param_len + strlen (SMB_LMAPI_SLOT) + 1 + data_len);

  p = (char*) SMB_Hdr (pkt) + SMB_trans_len + 2; /* Skip the BCC and ECT */

  strcpy (p, SMB_LMAPI_SLOT);
  p += strlen (SMB_LMAPI_SLOT) + 1;

  if (pad_api_name == TRUE)              /* Pad to a word boundary */
     p++;

  SSVAL (p, 0, SMB_LMapi_SetUserInfo);   /* The api call */
  p += 2;

  strcpy (p, SMB_LMAPI_SUI_DESC);        /* Copy in the param desc */
  p += strlen (SMB_LMAPI_SUI_DESC) + 1;

  strcpy (p, SMB_LMAPI_SUI_DATA_DESC);   /* Copy in second descriptor */
  p += strlen (SMB_LMAPI_SUI_DATA_DESC) + 1;

  strcpy (p, user);
  p += strlen (user) + 1;

  SSVAL (p, 0, 1);               /* Claim that we have a level 1 struct ? */
  p += 2;

  SSVAL (p, 0, 3);               /* Set the password */
  SSVAL (p, 2, 1);               /* Seems to be one ... */
  SSVAL (p, 4, strlen(newpass)); /* Length of new password ... */

  /* Now copy the data in ...
   */
  p += 6;
  if (pad_params == TRUE)
     p++;

  strcpy (p, newpass);

  /* Now send the lot and get a response ...
   */
  if (RFCNB_Send (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error sending Trans SetUserInfo request\n"));
    SMBlib_errno = SMBlibE_SendFailed;
    rc = -1;
    goto quit;
  }

  /* Now get the response ...
   */
  if (RFCNB_Recv (tree->con->transConnect, pkt, pkt_len) < 0)
  {
    SMB_DEBUG (1, ("Error receiving response to Trans SetUserInfo request\n"));
    SMBlib_errno = SMBlibE_RecvFailed;
    rc = -1;
    goto quit;
  }

  /* Check out the response type ...
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_DEBUG (1, ("SMB_trans SetUserInfo fail: errorclass = %d, Error Code = %d\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset)));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    SMBlib_errno = SMBlibE_Remote;
    rc = -1;
    goto quit;
  }

  /* All ok, pass back the status
   */
  *apiStatus = SVAL (SMB_Hdr(pkt), SVAL (SMB_Hdr(pkt), SMB_transr_pbo_offset));
  rc = 0;

quit:
  RFCNB_FREE_PKT (pkt);
  return (rc);
}

/*
 * List all the shares available on a server
 */
int SMBapi_NetShareEnum (struct SMB_Tree *tree,
                         char *enum_buf,
                         int   bufsiz,
                         int  *shares_returned,
                         int  *shares_total)
{
  ARGSUSED (enum_buf);
  ARGSUSED (bufsiz);
  ARGSUSED (shares_returned);
  ARGSUSED (shares_total);
  return (-1);
}


#if !defined(__DJGPP__) || (defined(__DJGPP__) && (__DJGPP_MINOR__ < 4))
void strlcpy (char *dst, const char *src, int len)
{
  SMB_ASSERT (len > 0);

  if (strlen(src) < len)
     strcpy (dst, src);
  else
  {
    memcpy (dst, src, len-1);
    dst [len-1] = '\0';
  }
}
#endif

/*
 * Figure out what protocol was accepted, given the list of dialect
 * strings we offered, and the index back from the server. We allow
 * for a user supplied list, and assume that it is a subset of our list
 */
static int FigureProtocol (const char **table, int prot_index)
{
  int i;

#if 0
  /* The jobs is easy, just index into table
   */
  if (table == SMB_protocols)
     return (SMB_protocols[prot_index].value);
#endif

  /* Search through table looking for a match
   */
  for (i = 0; SMB_protocols[i].name; i++)
      if (!strcmp (table[prot_index], SMB_protocols[i].name))
         return (SMB_protocols[i].value);
  return (-1);
}
