/*
 * Test the performance of SMB servers ... A start based on some ideas I had
 *
 * Version 1.0
 *
 * Copyright (C) Richard Sharpe 1996
 *
 */

/* This program allows you to read or write files on an SMB server and it
 * measures the performance of those reads and writes ...
 *
 * We take 10 (this number is settable from command line) measurements
 * and list them all so we can average.
 *
 * We read or write 10 * 1024 * 1024 bytes (my measure of a MB). If reading
 * we ensure the file exists first.
 *
 * We split the write up into 1K, 2K, 4K, 8K, 16K, and 32K chunks and measure
 * each lot. We have to take care to use a TCON against Windows NT 3.51.
 *
 * Command line is:
 *
 * -v means verbose (also Watt-32 debugging)
 * -l means logon, username and password needed ...
 * -u means user
 * -p means password
 * -r sets the repeat count
 * -s sets the file service to try against
 * -w means check write perf. Read is the default.
 * -n means set TCP_NODELAY on socket ...
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#ifndef _MSC_VER
#include <unistd.h>
#endif
#include <time.h>
#include <io.h>
#include <sys/stat.h>

#include <rfcnb.h>
#include <smblib.h>
#include <tcp.h>

#define TEST_REPEAT_COUNT   10
#define TEST_FILE_NAME      "\\testfile.dat"
#define TEST_USERNAME       "fred"
#define TEST_PASSWORD       "frednurk"
#define TEST_FILE_SERVICE   "msfax"
#define TEST_WRITE_SIZE     4096

int  local_io          = FALSE;
int  verbose           = FALSE;
int  logon             = FALSE;
int  test_write        = FALSE;
int  tcp_nodelay       = FALSE;
int  total_file_size   = 10 * 1024 * 1024;
int  write_chunk_size  = 0;
int  repeat_count      = 0;
const char **prots_to_use;

char username[80]      = "";
char password[80]      = "";
char file_service[128] = "";
char file_name[80]     = "";
char server_name[80]   = "";
char local_file_name[128];

void usage (char *prog)
{
  fprintf (stderr, "%s -l -v -u <username> -p <password> -r <count> -s", prog);
  fprintf (stderr, " <file-service> \\\n");
  fprintf (stderr, "          -n -w -i <local-file> -f <filename> -c <chunk size> <server>\n");
  fprintf (stderr, "\t-l means logon\n");
  fprintf (stderr, "\t-v means verbose\n");
  fprintf (stderr, "\t-u <username> provides the username for logon\n");
  fprintf (stderr, "\t-p <password> provides the password for logon\n");
  fprintf (stderr, "\t-r <count> provides the repeat count of tests to do\n");
  fprintf (stderr, "\t-s <service> provides the service to use\n");
  fprintf (stderr, "\t-f <filename> provides the file name to use\n");
  fprintf (stderr, "\t-c <chunk size> sets the chunk size\n");
  fprintf (stderr, "\t-n means set the TCP_NODELAY socket option\n");
  fprintf (stderr, "\t-i means read/write to/from local file\n");
  fprintf (stderr, "\t<server> provides the server to connect to.\n");
}

void do_error (int error)
{
  char err_str[1024];

  if (error == SMBlibE_Remote)
  {
    error = SMB_Get_Last_SMB_Err();
    SMB_Get_SMB_Error_Msg (SMBlib_Error_Class (error),
                           SMBlib_Error_Code (error),
                           err_str, sizeof(err_str)-1);
  }
  else
    SMB_Get_Error_Msg (error, err_str, sizeof(err_str)-1);

  printf ("  %s\n", err_str);
}

const char *SMB_prots[] = {
     "PC NETWORK PROGRAM 1.0",
     "MICROSOFT NETWORKS 1.03",
     "MICROSOFT NETWORKS 3.0",
     "LANMAN1.0",
     "LM1.2X002",
     "LANMAN2.1",
     "NT LM 0.12",
     "NT LANMAN 1.0",
     NULL
   };

struct SMB_Connect *connect_server (const char *server,
                                    const char *user,
                                    const char *password)
{
  struct SMB_Connect *con;
  int    prot;

  fprintf (stderr, "Connecting to server \\\\%s ...\n", server);

  con = SMB_Connect_Server (NULL, server);
  if (!con)
  {
    fprintf (stderr, "Unable to connect to server \"%s\" ... \n", server);
    do_error (SMB_Get_Last_Error());
    exit (1);
  }

  if (verbose)
     fprintf (stderr, "Connected to server \\\\%s ...\n", server);

  /* Now do a negotiate ... */

  if (verbose)
     fprintf (stderr, "Now we negotiate a protocol ... \n");


  if (SMB_Negotiate (con, SMB_prots) < 0)
  {
    fprintf (stderr, "Unable to negotiate a protocol of the offered.\n");
    do_error (SMB_Get_Last_Error());
    exit (1);
  }

  if (verbose)
  {
    prot = SMB_Get_Protocol_IDX (con);
    fprintf (stderr, "  The protocol accepted was %d, %s.\n",
             prot, SMB_prots[prot]);
  }

  /* Now we do a logon ... */

  if (logon)
  {
    if (verbose)
       fprintf (stderr, "Now try to logon as %s ...\n", user);

    if (SMB_Logon_Server (con, user, password) < 0)
    {
      fprintf (stderr, "Unable to logon to server ...\n");
      do_error (SMB_Get_Last_Error());
      fprintf (stderr, "This error ignored. We will keep going and hope ...\n");
    }
  }
  else
    fprintf (stderr, "We will not try to logon ...\n");

  return (con);
}

/*
 * Write a file ...
 */
void write_file (struct SMB_Tree *tree, const char *file_name, int local_fd)
{
  struct SMB_File *file;
  struct stat      file_stats;
  int    i, writes;
  time_t start_time, end_time;
  int    this_write, actual;
  char  *write_chunk;

  /* Now open the file requested for write, as we need to write it */

  if (verbose)
     fprintf (stderr, "Now we Create file: %s for writing.\n", file_name);

  file = SMB_Create (tree, NULL, file_name, SMB_FA_ORD);
  if (!file)
  {
    do_error (SMB_Get_Last_Error());
    exit (1);
  }

  if (verbose)
     fprintf (stderr, "File: %s created for writing...\n", file_name);

  /* Now we write the file */

  write_chunk = malloc (write_chunk_size);
  for (i = 0; i < write_chunk_size; i++)
      write_chunk[i] = i;

  if (local_io)
  {
    if (fstat(local_fd, &file_stats) < 0)
    {
      perror ("Could not get stats for file");
      exit (1);
    }
    total_file_size = file_stats.st_size; /* Hmmm if file too large ... */
  }

  writes = total_file_size / write_chunk_size;

  if (verbose)
     fprintf (stderr, "Total writes: %d, write chunk size: %d\n",
              writes, write_chunk_size);

  time (&start_time);

  while (total_file_size > 0)
  {
    if (total_file_size >= write_chunk_size)
         this_write = write_chunk_size;
    else this_write = total_file_size;

    if (local_io)
    {
      if (read(local_fd, write_chunk, this_write) < 0)
      {
        perror ("Could not read file");
        exit (1);
      }
    }

    actual = SMB_Write (file, write_chunk, this_write);
    if (actual < 0)
    {
      do_error (SMB_Get_Last_Error());
      exit (1);
    }

    if (actual < this_write)
       fprintf (stderr, "Actual write was smaller than asked for!\n");

    total_file_size -= this_write;
  }

  time (&end_time);

  fprintf (stderr, "Total time to write: %d seconds\n", end_time - start_time);
#if 0
  fprintf (stderr, "Total time: %d clock ticks.\n", t2 - t1);
#endif
}

void read_file (const void *tree, const char *file_name)
{
}

int main (int argc, char **argv)
{
  struct SMB_Connect *con;
  struct SMB_Tree    *tree;
  int    opt, local_fd;
  char   file_name_tmp[128];
  char  *prog =   argv[0];

  write_chunk_size = TEST_WRITE_SIZE;

  while ((opt = getopt (argc, argv, "mlnvwf:i:p:u:s:r:c:")) != EOF)
  {
    switch (opt)
    {
      case 'i':             /* Set up a local file */
           local_io = TRUE;
           strncpy (local_file_name, optarg, sizeof (local_file_name) - 1);
           break;

      case 'n':
           tcp_nodelay = TRUE;
           break;

      case 'c':
           write_chunk_size = atoi (optarg);
           break;

      case 'l':             /* do a logon and tree connect ... */
           logon = TRUE;
           break;

      case 'v':             /* Verbose? */
           verbose = TRUE;
           dbug_init();
           break;

      case 'w':
           test_write = TRUE;
           break;

      case 'f':             /* Get file name to use */
           strncpy (file_name, optarg, sizeof(file_name) - 1);
           break;

      case 'p':             /* Get password */
           strncpy (password, optarg, sizeof(password) - 1);
           break;

      case 'u':             /* Pick up the user name */
           strncpy (username, optarg, sizeof(username) - 1);
           break;

      case 's':
           strncpy (file_service, optarg, sizeof(file_service) - 1);
           break;

      case 'r':
           repeat_count = atoi (optarg);
           break;

      default:
           usage (prog);
           exit (1);
    }
  }

  if (optind < argc)
     strncpy (server_name, argv[optind], sizeof(server_name)-1);
  else
  {
    usage (prog);
    exit (1);
  }

  /* Check parameters */

  if (repeat_count <= 0)
     repeat_count = TEST_REPEAT_COUNT;

  if (file_name[0] == 0)
     strncpy (file_name, TEST_FILE_NAME, sizeof (file_name) - 1);

  if (file_name[0] != '\\')
  {
    strncpy (file_name_tmp, file_name, sizeof(file_name_tmp) - 1);
    strncpy (file_name, "\\", sizeof(file_name) - 1);
    strncat (file_name, file_name_tmp, sizeof(file_name) - strlen(file_name)-1);
  }

  if (username[0] == 0)
     strncpy (username, TEST_USERNAME, sizeof(username) - 1);

  if (password[0] == 0)
     strncpy (password, TEST_PASSWORD, sizeof(password) - 1);

  if (file_service[0] == 0)
     strncpy (file_service, TEST_FILE_SERVICE, sizeof(file_service) - 1);

  /* Go to it ... */

  if (verbose)
  {
    fprintf (stderr, "User: %s \nPassword: %s \nFile Service: %s \n",
             username, password, file_service);

    fprintf (stderr, "Server: %s\nFilename: %s\nRepeat count: %d\n\n",
             server_name, file_name, repeat_count);
  }

  /* Well, now we have the parameters we want, we go to it
   */
  if (local_io)
  {
    local_fd = open (local_file_name, O_RDWR, 0);
    if (local_fd < 0)
    {
      fprintf (stderr, "File: %s could not be opened\n", local_file_name);
      perror ("");
      exit (1);
    }
  }

  SMB_Init();

  con = connect_server (server_name, username, password);

  /* Now connect to the file serice */

  strncpy (file_name_tmp, file_service, sizeof (file_name_tmp) - 1);
  strncpy (file_service, "\\\\", sizeof (file_name_tmp) - 1);
  strncat (file_service, server_name, sizeof (file_service) - strlen (file_service) - 1);
  strncat (file_service, "\\", sizeof (file_service) - strlen (file_service) - 1);
  strncat (file_service, file_name_tmp, sizeof (file_service) - strlen (file_service) - 1);

  if (verbose)
     fprintf (stderr, "Now we connect to service: %s\n", file_service);

  if (tcp_nodelay)
     SMBlib_Set_Sock_NoDelay (con, TRUE);

  tree = SMB_TreeConnect (con, NULL, file_service, "", "A:");
  if (!tree)
  {
    do_error (SMB_Get_Last_Error());
    exit (1);
  }

  if (verbose)
  {
    fprintf (stderr, "Connected to service: %s\n", file_service);
    fprintf (stderr, "The Max Buffer Size is: %d\n", SMB_Get_Tree_MBS (tree));
  }

  if (test_write)
     write_file (tree, file_name, local_fd);
  else
  {
    /* Read, but we create the file first, then read it ... */
    write_file (tree, file_name, local_fd);

    /* Now loop through reading the file ... and time it ... */
    read_file (tree, file_name);
  }

  if (verbose)
     fprintf (stderr, "All done, disconnecting...\n");

  SMB_Discon (&con, FALSE);
  return (0);
}

